﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;

namespace M64_to_MIDI
{
    public partial class main : Form
    {
        public main()
        {
            InitializeComponent();
            ChunkSelection.SelectedIndex = 0;
        }

        public class SettingsData
        {
            public byte LimitChunks = new byte();
            public bool Restrict = new bool();
        }
        public static SettingsData Settings = new SettingsData();

        public class HeaderClass
        {
            public bool[,] ActiveLayers = new bool[0x10, 0x10];
            public Int64 ReturnOffset = new Int64();
            public Int32 TotalTimestamp = new Int32();
            public Int32 MaxTimestamp = new Int32();
            public Int32 LoopOffset = new Int32();
            public Int16 TotalLayers = new Int16();
            public byte Volume = new byte();
            public byte Tempo = new byte();
            public bool Valid = false;
        }
        public static HeaderClass HeaderData = new HeaderClass();

        public class TrackClass
        {
            public Int64 ReturnOffset = new Int64();
            public Int32 TotalTimestamp = new Int32();
            public byte Program = new byte();
            public byte Pan = new byte();
        }
        public static TrackClass TrackData = new TrackClass();

        public class LayerClass
        {
            public Int64 ReturnOffset = new Int64();
            public Int64 WriteOffset = new Int64();
            public Int32 TotalTimestamp = new Int32();
        }
        public static LayerClass LayerData = new LayerClass();

        public class NoteClass
        {
            public Int32 Timestamp = new Int32();
            public Int32 RestTimestamp = new Int32();
            public byte Transposition = new byte();
            public byte Key = new byte();
            public byte Velocity = new byte();
            public byte Duration = new byte();
            public bool FirstNote = true;
            public bool InJump = false;
        }
        public static NoteClass NoteData = new NoteClass();

        public class Reverse
        {
            public static Int16 Int16(Int16 value)
        {
            Byte[] bytes = BitConverter.GetBytes(value);
            Array.Reverse(bytes);
            return BitConverter.ToInt16(bytes, 0);
        }

            public static Int32 Int32(Int32 value)
        {
            Byte[] bytes = BitConverter.GetBytes(value);
            Array.Reverse(bytes);
            return BitConverter.ToInt32(bytes, 0);
        }

            public static Int32 String32(string text)
            {
                Int32 value = 0x0000;
                char[] array = text.ToCharArray();
                foreach (char c in array) value = (value << 8) + c;
                Byte[] bytes = BitConverter.GetBytes(value);
                Array.Reverse(bytes);
                return BitConverter.ToInt32(bytes, 0);
            }
        }

        public static byte FirstByte(Int32 value)
        {
            Byte[] bytes = BitConverter.GetBytes(value);
            return bytes[0];
        }

        public static void Advance(BinaryReader br, Int32 value)
        {
            br.BaseStream.Position += value;
        }

        public static void ReadHeader(BinaryReader br)
        {
            while (br.BaseStream.Position < br.BaseStream.Length)
            {
                byte by = br.ReadByte();
                if ((by >= 0x90) && (by <= 0x9F)) Advance(br, 2);
                if (by == 0xC4) break;
                if (by == 0xC8) Advance(br, 1);
                if (by == 0xD3)
                {
                    Advance(br, 1);
                    HeaderData.Valid = true;
                }
                if (by == 0xD5) Advance(br, 1);
                if (by == 0xD6) Advance(br, 2);
                if (by == 0xD7) Advance(br, 2);
                if (by == 0xD9) Advance(br, 1);
                if (by == 0xDB) HeaderData.Volume = br.ReadByte();
                if (by == 0xDD) HeaderData.Tempo = br.ReadByte();
                if (by == 0xF9) Advance(br, 2);
                if (by == 0xFA) Advance(br, 2);
                if (by == 0xFB) HeaderData.LoopOffset = Reverse.Int16(br.ReadInt16()) & 0xFFFF;
                if (by == 0xFD) HeaderData.MaxTimestamp += ReadParameter(br);
                if (by == 0xFF) break;
            }
        }

        public static void ReadNote(BinaryReader br, byte by)
        {
            Int32 key = new Int32();
            if ((by >= 0x00) && (by <= 0x3F))
            {
                key = by;
                NoteData.Timestamp = ReadParameter(br);
                NoteData.Velocity = br.ReadByte();
                NoteData.Duration = br.ReadByte();
            }
            if ((by >= 0x40) && (by <= 0x7F))
            {
                key = by - 0x40;
                NoteData.Timestamp = ReadParameter(br);
                NoteData.Velocity = br.ReadByte();
            }
            if ((by >= 0x80) && (by <= 0xBF))
            {
                key = by - 0x80;
                NoteData.Velocity = br.ReadByte();
                NoteData.Duration = br.ReadByte();
            }
            NoteData.Key = FirstByte(key);
            LayerData.TotalTimestamp += NoteData.Timestamp;
        }

        public static Int16 ReadParameter(BinaryReader br)
        {
            Int32 value = br.ReadByte();
            if ((value & 0x80) == 0x80)
            {
                value = (value << 8) + br.ReadByte();
                if ((value & 0x7F80) == 0x00)
                {
                    value = 0xC0;
                    br.BaseStream.Position--;
                }
            }

            Byte[] bytes = BitConverter.GetBytes(value & 0x7FFF);
            Int16 parameter = BitConverter.ToInt16(bytes, 0);
            return parameter;
        }

        public static void ConvertHeader(BinaryReader br, BinaryWriter bw, byte track, byte layer)
        {
            if (Settings.LimitChunks == 2) br.BaseStream.Position = HeaderData.LoopOffset;
            while (br.BaseStream.Position < br.BaseStream.Length)
            {
                byte by = br.ReadByte();
                if ((by >= 0x90) && (by <= 0x9F))
                {
                    if (by == (0x90 | track))
                    {
                        if (HeaderData.TotalTimestamp > LayerData.TotalTimestamp) NoteData.RestTimestamp = HeaderData.TotalTimestamp - LayerData.TotalTimestamp;
                        HeaderData.ReturnOffset = br.BaseStream.Position + 0x02;
                        br.BaseStream.Position = Reverse.Int16(br.ReadInt16()) & 0xFFFF;
                        ConvertTrack(br, bw, track, layer);
                        br.BaseStream.Position = HeaderData.ReturnOffset;
                    }
                    else Advance(br, 2);
                }
                if (by == 0xC4) break;
                if (by == 0xC8) Advance(br, 1);
                if (by == 0xD3) Advance(br, 1);
                if (by == 0xD5) Advance(br, 1);
                if (by == 0xD6) Advance(br, 2);
                if (by == 0xD7) Advance(br, 2);
                if (by == 0xD9) Advance(br, 1);
                if (by == 0xDB) Advance(br, 1);
                if (by == 0xDD) Advance(br, 1);
                if (by == 0xF9) Advance(br, 2);
                if (by == 0xFA) Advance(br, 2);
                if (by == 0xFB) Advance(br, 2);
                if (by == 0xFD) HeaderData.TotalTimestamp += ReadParameter(br);
                if (by == 0xFF) break;
                if ((Settings.LimitChunks == 1) && (br.BaseStream.Position >= HeaderData.LoopOffset)) break;
            }
        }

        public static void ConvertTrack(BinaryReader br, BinaryWriter bw, byte track, byte layer)
        {
            while (br.BaseStream.Position < br.BaseStream.Length)
            {
                byte by = br.ReadByte();
                if ((by >= 0x90) && (by <= 0x9F))
                {
                    if (by == (0x90 | layer))
                    {
                        TrackData.ReturnOffset = br.BaseStream.Position + 0x02;
                        br.BaseStream.Position = Reverse.Int16(br.ReadInt16()) & 0xFFFF;
                        ConvertLayer(br, bw, track, layer);
                        br.BaseStream.Position = TrackData.ReturnOffset;

                        if (!HeaderData.ActiveLayers[track, by & 0x0F]) HeaderData.TotalLayers++;
                        HeaderData.ActiveLayers[track, by & 0x0F] = true;
                    }
                    else Advance(br, 2);
                }
                if (by == 0xC1) TrackData.Program = br.ReadByte();
                if (by == 0xC2) Advance(br, 2);
                if (by == 0xC6) Advance(br, 1);
                if (by == 0xC8) Advance(br, 1);
                if (by == 0xCA) Advance(br, 1);
                if (by == 0xCC) Advance(br, 2);
                if (by == 0xD0) Advance(br, 1);
                if (by == 0xD3) Advance(br, 1);
                if (by == 0xD4) Advance(br, 1);
                if (by == 0xD8) Advance(br, 1);
                if (by == 0xDC) Advance(br, 1);
                if (by == 0xDD) TrackData.Pan = br.ReadByte();
                if (by == 0xDF) Advance(br, 1);
                if (by == 0xE0) Advance(br, 1);
                if (by == 0xE5) Advance(br, 1);
                if (by == 0xE9) Advance(br, 1);
                if (by == 0xFA) Advance(br, 2);
                if (by == 0xFB) Advance(br, 2);
                if (by == 0xFD) TrackData.TotalTimestamp += ReadParameter(br);
                if (by == 0xFE) Advance(br, 1);
                if (by == 0xFF) break;
            }
        }

        public static void ConvertLayer(BinaryReader br, BinaryWriter bw, byte track, byte layer)
        {
            NoteData.FirstNote = true;
            while (br.BaseStream.Position < br.BaseStream.Length)
            {
                byte by = br.ReadByte();
                if ((by >= 0x00) && (by <= 0xBF))
                {
                    ReadNote(br, by);
                    WriteNote(bw, track);
                    NoteData.RestTimestamp = 0;
                    NoteData.FirstNote = false;
                }
                if (by == 0xC0)
                {
                    Int16 value = ReadParameter(br);
                    NoteData.RestTimestamp += value;
                    LayerData.TotalTimestamp += value;
                }
                if (by == 0xC2) NoteData.Transposition = br.ReadByte();
                if (by == 0xFC)
                {
                    LayerData.ReturnOffset = br.BaseStream.Position + 0x02;
                    br.BaseStream.Position = Reverse.Int16(br.ReadInt16()) & 0xFFFF;
                    NoteData.InJump = true;
                }
                if (by == 0xFF)
                {
                    if (!NoteData.InJump) break;
                    br.BaseStream.Position = LayerData.ReturnOffset;
                    NoteData.InJump = false;
                }
                if ((LayerData.TotalTimestamp >= HeaderData.MaxTimestamp) && Settings.Restrict) break;
            }
        }

        public class WriteControl
        {
            public static void Pan(BinaryWriter bw, byte track, Int32 delta, byte pan)
            {
                WriteDelta(bw, delta);
                bw.Write(FirstByte(0xB0 | track));
                bw.Write(FirstByte(0x0A));
                bw.Write(pan);
            }
            public static void Volume(BinaryWriter bw, byte track, Int32 delta, byte volume)
            {
                WriteDelta(bw, delta);
                bw.Write(FirstByte(0xB0 | track));
                bw.Write(FirstByte(0x07));
                if (volume > 0x7F) volume = 0x7F;
                bw.Write(volume);
            }
            public static void Program(BinaryWriter bw, byte track, Int32 delta, byte program)
            {
                WriteDelta(bw, delta);
                bw.Write(FirstByte(0xC0 | track));
                bw.Write(program);
            }
        }

        public static void WriteMeta(BinaryWriter bw, Int32 delta, Int32 value)
        {
            WriteDelta(bw, delta);
            bw.Write(FirstByte(0xFF));
            bw.Write(FirstByte(value));
        }

        public static void WriteDelta(BinaryWriter bw, Int32 delta)
        {
            Int32 converted = new Int32();
            for (byte i = new byte(); i < 0x04; i++)
            {
                byte by = FirstByte((delta << i) >> (8 * i));
                if (i == 0x00) converted = converted | ((by & 0x7F) << (8 * i));
                else if (by != 0x00) converted = converted | (((by | 0x80) & 0xFF) << (8 * i));
            }

            Byte[] bytes = BitConverter.GetBytes(converted);
            if (bytes[3] != 0x00) bw.Write(Reverse.Int32(BitConverter.ToInt32(bytes, 0)));
            else if (bytes[2] != 0x00)
            {
                bw.Write(bytes[2]);
                bw.Write(bytes[1]);
                bw.Write(bytes[0]);
            }
            else if (bytes[1] != 0x00) bw.Write(Reverse.Int16(BitConverter.ToInt16(bytes, 0)));
            else bw.Write(bytes[0]);
        }

        public static void WriteNote(BinaryWriter bw, byte track)
        {
            WriteDelta(bw, NoteData.RestTimestamp);

            if (NoteData.FirstNote) bw.Write(FirstByte(0x90 | track));

            bw.Write(FirstByte(NoteData.Key + NoteData.Transposition + 0x15));
            bw.Write(NoteData.Velocity);

            WriteDelta(bw, NoteData.Timestamp);

            bw.Write(FirstByte(NoteData.Key + NoteData.Transposition + 0x15));
            bw.Write(FirstByte(0x00));
        }

        public static void StartMIDIHeader(BinaryWriter bw)
        {
            // Chunk ID.
            bw.Write(Reverse.String32("MThd"));

            // Header length.
            bw.Write(Reverse.Int32(0x00000006));

            // File format.
            bw.Write(Reverse.Int16(0x0001));

            // Reserve number of tracks.
            bw.Write(Reverse.Int16(0x0000));

            //PPQN clock.
            bw.Write(Reverse.Int16(0x0030)); 
        }

        public static void FinishMIDIHeader(BinaryWriter bw)
        {
            bw.BaseStream.Position = 0x0A;
            bw.Write(Reverse.Int16(++HeaderData.TotalLayers));
        }

        public static void StartMIDISettings(BinaryWriter bw)
        {
            // Chunk ID.
            bw.Write(Reverse.String32("MTrk"));

            // Header length.
            bw.Write(Reverse.Int32(0x00000019));

            // Time signature.
            WriteMeta(bw, 0x00, 0x58);
            bw.Write(FirstByte(0x04));
            bw.Write(FirstByte(0x04));
            bw.Write(FirstByte(0x02));
            bw.Write(FirstByte(0x18));
            bw.Write(FirstByte(0x08));

            // Key signature.
            WriteMeta(bw, 0x00, 0x59);
            bw.Write(FirstByte(0x02));
            bw.Write(FirstByte(0x00));
            bw.Write(FirstByte(0x00));

            // Reserve Tempo.
            WriteMeta(bw, 0x00, 0x51);
            bw.Write(Reverse.Int32(0x00000000));

            // End track.
            bw.Write(Reverse.Int32(0x00FF2F00));
        }

        public static void FinishMIDISettings(BinaryWriter bw)
        {
            bw.BaseStream.Position = 0x27;
            bw.Write(Reverse.Int32((60000000 / (HeaderData.Tempo & 0xFF)) | 0x03000000));
        }

        public static void StartMIDITrack(BinaryWriter bw, byte track)
        {
            // Chunk ID.
            bw.Write(Reverse.String32("MTrk"));

            // Reserve track length.
            bw.Write(Reverse.Int32(0x00000000));
            LayerData.WriteOffset = bw.BaseStream.Position;

            // Reserve track name.
            WriteMeta(bw, 0x00, 0x03);
            bw.Write(FirstByte(0x00));

            // Reserve program.
            WriteControl.Program(bw, track, 0x00, 0x00);

            // Volume.
            WriteControl.Volume(bw, track, 0x00, 0x7F);

            // Reserve pan.
            WriteControl.Pan(bw, track, 0x00, 0x00);
        }

        public static void FinishMIDITrack(BinaryWriter bw, byte track)
        {
            bw.Write(Reverse.Int32(0x00FF2F00));
            Int64 length = bw.BaseStream.Position - LayerData.WriteOffset;
            bw.BaseStream.Position = LayerData.WriteOffset - 0x04;
            Byte[] bytes = BitConverter.GetBytes(length);
            bw.Write(Reverse.Int32(BitConverter.ToInt32(bytes, 0)));

            // program.
            bw.BaseStream.Position = LayerData.WriteOffset + 0x04;
            WriteControl.Program(bw, track, 0x00, TrackData.Program);

            // pan.
            bw.BaseStream.Position = LayerData.WriteOffset + 0x0B;
            WriteControl.Pan(bw, track, 0x00, TrackData.Pan);

            if (LayerData.TotalTimestamp > 0) bw.BaseStream.Position = bw.BaseStream.Length;
            else bw.BaseStream.Position = bw.BaseStream.Length - 0x1B;
        }

        private void open_Click(object sender, EventArgs e)
        {
            OpenFileDialog dlg = new OpenFileDialog();
            dlg.Title = "Open sequence file";
            dlg.Filter = "All Files|*.*";
            if (dlg.ShowDialog() == DialogResult.OK) file.Text = dlg.FileName;
            dlg.Dispose();
        }

        private void convert_Click_1(object sender, EventArgs e)
        {
            // Setup.
            if (!File.Exists(file.Text))
            {
                MessageBox.Show("Load a valid M64 file first!", "Error!", 0, MessageBoxIcon.Error);
                return;
            }

            string extension = ".mid";
            if (Settings.LimitChunks == 1) extension = ".0.mid";
            if (Settings.LimitChunks == 2) extension = ".1.mid";

            File.Delete(file.Text + extension);
            FileStream fileIn = new FileStream(file.Text, FileMode.Open, FileAccess.Read);
            FileStream fileOut = File.Create(file.Text + extension);

            BinaryReader br = new BinaryReader(fileIn);
            BinaryWriter bw = new BinaryWriter(fileOut);

            // Check valid M64.
            ReadHeader(br);
            if (!HeaderData.Valid)
            {
                MessageBox.Show("This is not a valid M64 file!", "Error!", 0, MessageBoxIcon.Error);
                return;
            }

            // Start MIDI header.
            StartMIDIHeader(bw);

            // Start MIDI settings.
            StartMIDISettings(bw);

            // Convert M64 data.
            for (byte track = new byte(); track < 0x10; track++)
            {
                for (byte layer = new byte(); layer < 0x10; layer++)
                {
                    HeaderData.TotalTimestamp = 0;
                    br.BaseStream.Position = 0;

                    StartMIDITrack(bw, track);
                    ConvertHeader(br, bw, track, layer);
                    FinishMIDITrack(bw, track);

                    TrackData = new TrackClass();
                    LayerData = new LayerClass();
                    NoteData = new NoteClass();
                }
            }

            // Finish MIDI header.
            FinishMIDIHeader(bw);

            // Finish MIDI settings.
            FinishMIDISettings(bw);

            // Finish.
            br.Close();
            bw.Close();

            fileIn.Close();
            fileOut.Close();

            HeaderData = new HeaderClass();

            MessageBox.Show("Conversion to MIDI complete!", "Done!", 0, MessageBoxIcon.None);
        }

        private void about_Click(object sender, EventArgs e)
        {
            MessageBox.Show("This program interprets the music format used in many first-party N64 games \nsuch as Super Mario 64, Mario Kart 64 or Star Fox 64, and outputs a MIDI file.", "About", 0, MessageBoxIcon.None);
        }

        private void restrict_CheckedChanged(object sender, EventArgs e)
        {
            Settings.Restrict = Restrict.Checked;
        }

        private void ChunkSelection_SelectedIndexChanged(object sender, EventArgs e)
        {
            Settings.LimitChunks = FirstByte(ChunkSelection.SelectedIndex);
        }
    }
}
